package dev.fyloz.colorrecipesexplorer.config.security.filters

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.logic.account.JwtLogic
import dev.fyloz.colorrecipesexplorer.logic.account.UserJwt
import io.jsonwebtoken.ExpiredJwtException
import org.springframework.security.authentication.AuthenticationManager
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken
import org.springframework.security.core.Authentication
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.security.web.authentication.www.BasicAuthenticationFilter
import org.springframework.web.util.WebUtils
import javax.servlet.FilterChain
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse

class JwtAuthorizationFilter(
    private val jwtLogic: JwtLogic,
    authenticationManager: AuthenticationManager
) : BasicAuthenticationFilter(authenticationManager) {
    override fun doFilterInternal(request: HttpServletRequest, response: HttpServletResponse, chain: FilterChain) {
        val authorizationCookie = WebUtils.getCookie(request, Constants.HeaderNames.AUTHORIZATION)

        // If there is no authorization cookie, the user is not authenticated
        if (authorizationCookie == null) {
            chain.doFilter(request, response)
            return
        }

        val authorizationToken = authorizationCookie.value
        if (!isJwtValid(authorizationToken)) {
            chain.doFilter(request, response)
            return
        }

        SecurityContextHolder.getContext().authentication = getAuthentication(authorizationToken)
        chain.doFilter(request, response)
    }

    // The authorization token is valid if it starts with "Bearer"
    private fun isJwtValid(authorizationToken: String) =
        authorizationToken.startsWith(JwtAuthenticationFilter.BEARER_TOKEN_PREFIX)

    private fun getAuthentication(authorizationToken: String): Authentication? {
        return try {
            val jwt = authorizationToken.replace(JwtAuthenticationFilter.BEARER_TOKEN_PREFIX, "").trim()
            val user = jwtLogic.parseJwt(jwt)

            getAuthentication(user)
        } catch (_: ExpiredJwtException) {
            null
        }
    }

    private fun getAuthentication(user: UserJwt) =
        UsernamePasswordAuthenticationToken(user.id, null, user.authorities)
}
