package dev.fyloz.colorrecipesexplorer.logic.account

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.dtos.account.UserDetails
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.model.account.toAuthority
import dev.fyloz.colorrecipesexplorer.utils.base64encode
import dev.fyloz.colorrecipesexplorer.utils.toDate
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.jackson.io.JacksonDeserializer
import io.jsonwebtoken.jackson.io.JacksonSerializer
import org.springframework.security.core.GrantedAuthority
import org.springframework.stereotype.Service
import java.time.Instant
import java.util.*

const val jwtClaimUser = "user"

interface JwtLogic {
    /** Build a JWT token for the given [userDetails]. */
    fun buildJwt(userDetails: UserDetails): String

    /** Parses a user information from the given [jwt] token. */
    fun parseJwt(jwt: String): UserJwt
}

@Service
class DefaultJwtLogic(
    val objectMapper: ObjectMapper,
    val securityProperties: CreSecurityProperties
) : JwtLogic {
    private val secretKey by lazy {
        securityProperties.jwtSecret.base64encode()
    }

    private val jwtBuilder by lazy {
        Jwts.builder()
            .serializeToJsonWith(JacksonSerializer<Map<String, *>>(objectMapper))
            .signWith(secretKey)
    }

    private val jwtParser by lazy {
        Jwts.parserBuilder()
            .deserializeJsonWith(JacksonDeserializer<Map<String, *>>(objectMapper))
            .setSigningKey(secretKey)
            .build()
    }

    override fun buildJwt(userDetails: UserDetails): String =
        jwtBuilder
            .setSubject(userDetails.id)
            .setExpiration(getCurrentExpirationDate())
            .claim(JWT_CLAIM_PERMISSIONS, objectMapper.writeValueAsString(userDetails.permissions))
            .compact()

    override fun parseJwt(jwt: String): UserJwt {
        val parsedJwt = jwtParser.parseClaimsJws(jwt)

        val serializedPermissions = parsedJwt.body.get(JWT_CLAIM_PERMISSIONS, String::class.java)
        val permissions = objectMapper.readValue<Collection<Permission>>(serializedPermissions)

        val authorities = permissions
            .map { it.toAuthority() }
            .toMutableList()

        return UserJwt(parsedJwt.body.subject, authorities)
    }

    private fun getCurrentExpirationDate(): Date =
        Instant.now()
            .plusSeconds(securityProperties.jwtDuration)
            .toDate()

    companion object {
        private const val JWT_CLAIM_PERMISSIONS = "permissions"
    }
}

data class UserJwt(val id: String, val authorities: Collection<GrantedAuthority>)